/*
 * Name:    main.cpp
 *
 * Purpose: This is the main routine for the MX CA server.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2009, 2011-2014 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <iostream>

#include "mxcaServer.h"

#include "fdManager.h"

extern "C" {
#include "mx_unistd.h"		/* For getopt() */
}

int
main( int argc, char **argv )
{
	mxcaServer *server;
	MX_RECORD *mx_record_list;
	MX_LIST_HEAD *mx_list_head;
	MX_PIPE *mx_callback_pipe;
	MX_CALLBACK_MESSAGE *mx_callback_message;
	MX_VIRTUAL_TIMER *mx_callback_timer;
	size_t num_bytes_available, bytes_read;
	int c, error_flag, debug_level;
	unsigned long mx_network_debug_flags;
	unsigned long epics_network_debug_flags;
	double wait_seconds;
	unsigned long wait_msec;
	mx_bool_type enable_remote_breakpoint = FALSE;
	mx_bool_type list_pvs = FALSE;
	mx_bool_type show_permission_tests = FALSE;
	mx_bool_type start_debugger = FALSE;
	mx_bool_type warn_missing_process_functions = FALSE;
	char *database_name;
	char *acl_name;
	unsigned long mxca_server_flags = 0;
	static const char usage[] =
"\n"
"Usage: mxca_server [options ...] 'mx_database_file' 'acl_file'\n"
"\n"
"  Options are:\n"
"    -a              Enable MX network client debugging\n"
"    -A              Verbose MX network client debugging\n"
"    -d debug_level  Set MX debugging level\n"
"    -D              Start debugger\n"
"    -e              Enable EPICS network server debugging\n"
"    -l              List created EPICS PVs\n"
"    -p              Show permission tests\n"
"    -r              Enable remote breakpoint\n"
"    -s scan_period  Set server scan period in seconds\n"
"    -w              Warn about missing MX process functions\n"
"    -W seconds      Wait for the specified number of seconds at startup\n"
"\n";

	mx_status_type mx_status;
	error_flag = FALSE;
	debug_level = 0;
	mx_network_debug_flags = 0;
	epics_network_debug_flags = 0;
	wait_seconds = 0.0;

	double scan_period = 1.0;	/* in seconds */
	
	while ((c = getopt(argc, argv, "aAd:Delprs:wW:")) != -1 )
	{
		switch(c) {
		case 'a':
			mx_network_debug_flags |= MXF_NETDBG_SUMMARY;
			break;
		case 'A':
			mx_network_debug_flags |= MXF_NETDBG_VERBOSE;
			break;
		case 'd':
			debug_level = atoi( optarg );
			break;
		case 'D':
			start_debugger = TRUE;
			break;
		case 'e':
			epics_network_debug_flags = TRUE;
			break;
		case 'l':
			list_pvs = TRUE;
			break;
		case 'p':
			show_permission_tests = TRUE;
			break;
		case 'r':
			enable_remote_breakpoint = TRUE;
			break;
		case 's':
			scan_period = atof( optarg );
			break;
		case 'w':
			warn_missing_process_functions = TRUE;
			break;
		case 'W':
			wait_seconds = atof( optarg );
			break;
		case '?':
			error_flag = TRUE;
			break;
		}
	}

	if ( argc < 3 ) {
		fputs( usage, stderr );
		exit(1);
	}

	if ( error_flag == TRUE ) {
		fputs( usage, stderr );
		exit(1);
	}

	database_name = argv[optind];
	acl_name = argv[optind+1];

	mx_set_debug_level(debug_level);

	if ( start_debugger ) {
		mx_start_debugger(NULL);
	}

	if ( wait_seconds >= 0.001 ) {
		mx_info( "Waiting for %f seconds", wait_seconds );

		wait_msec = mx_round( 1000.0 * wait_seconds );
		mx_msleep( wait_msec );
	}

	mxca_server_flags = 0;

	if ( show_permission_tests ) {
		mxca_server_flags |= MXCASF_SHOW_PERMISSION_TESTS;
	}

	/* Register a callback for MX export directives. */

	mx_status = mx_register_export_callback( (char *) "epics",
						mxca_export_callback_shim );

	if ( mx_status.code != MXE_SUCCESS )
		exit(mx_status.code);

	/* Create a channel access server. */

	server = new mxcaServer( database_name,
				scan_period,
				mxca_server_flags,
				mx_network_debug_flags,
				epics_network_debug_flags );

	if ( server == NULL ) {
		fprintf(stderr, "Could not create MX CA server.\n" );
		exit(1);
	}

	if ( list_pvs ) {
		/* Print out the names of the created PVs. */

		mxcaPV *PVList_head, *current_PV;

		PVList_head = server->PVList;

		current_PV = PVList_head;

		do {
			mx_info( "PV '%s'", current_PV->getName() );

			current_PV = current_PV->get_next_pv();

		} while ( current_PV != PVList_head );
	}

	/* Check to see if any of the PVs with MX record fields
	 * do not have a process function.
	 */

	if ( warn_missing_process_functions ) {
		mxcaPV *PVList_head, *current_PV;
		MX_RECORD_FIELD *field;

		PVList_head = server->PVList;

		current_PV = PVList_head;

		do {
			field = current_PV->get_mx_field();

			if ( field != (MX_RECORD_FIELD *) NULL ) {
				if ( field->process_function == NULL ) {
					mx_warning(
				"Local MX field '%s.%s' used by PV '%s' "
				"does not have a process function.",
					field->record->name,
					field->name,
					current_PV->getName() );
				}
			}

			current_PV = current_PV->get_next_pv();

		} while ( current_PV != PVList_head );
	}

	/* Initialize the socket connection access control list. */

	if ( strlen( acl_name ) > 0 ) {
		mx_status = mx_setup_connection_acl( server->mx_record_list,
							acl_name );

		if ( mx_status.code != MXE_SUCCESS )
			exit(mx_status.code);
	}

	/*--- Get a pointer to the MX callback pipe. ---*/

	mx_record_list = server->mx_record_list;

	mx_list_head = mx_get_record_list_head_struct( mx_record_list );

	mx_callback_pipe = (MX_PIPE *) mx_list_head->callback_pipe;

	mx_callback_timer = (MX_VIRTUAL_TIMER *) mx_list_head->callback_timer;

	/*--- Enable remote breakpoints (if requested) ---*/

	if ( enable_remote_breakpoint ) {
		mx_list_head->remote_breakpoint_enabled = TRUE;
	} else {
		mx_list_head->remote_breakpoint_enabled = FALSE;
	}

	/*--- Start the server scan timer. ---*/

	server->start();

	mx_info( "*** MX channel access server ready. ***" );

	/*--- Loop forever ---*/

	double delay = 0.01;

	while (aitTrue) {
		fileDescriptorManager.process( delay );

		mx_status = mx_pipe_num_bytes_available( mx_callback_pipe,
							&num_bytes_available );

		if ( (mx_status.code == MXE_SUCCESS)
		  && (num_bytes_available > 0) )
		{
			mx_status = mx_pipe_read( mx_callback_pipe,
						(char *) &mx_callback_message,
						sizeof(MX_CALLBACK_MESSAGE *),
						&bytes_read );

			if ( mx_callback_message->callback_type == MXCBT_POLL )
			{
				server->scan( mx_callback_timer );
			}
		}
	}

	return 1;
}

