/*
 * Name:    mxcaChannel.cpp
 *
 * Purpose: This file defines the mxcaChannel class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2009, 2011 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_CHANNEL_DEBUG	FALSE

#include <stdio.h>

#include "mxcaServer.h"

mxcaChannel::mxcaChannel( const casCtx &ctx,
			const char * const user,
			const char * const host,
			mxcaPV *channel_pv ) : casChannel( ctx )
{
	static const char fname[] = "mxcaChannel::mxcaChannel()";

	mxcaServer *mxca_server;
	MX_RECORD *mx_record_list;
	MX_LIST_HEAD *list_head;
	mx_bool_type connection_allowed;
	mx_bool_type *key_value;
	void *table_entry;
	unsigned long flags;
	mx_bool_type perform_dns_lookup;
	mx_status_type mx_status;

	this->username = user;
	this->hostname = host;
	this->pv = channel_pv;

	flags = channel_pv->epicsServer->mxca_server_flags;

	if ( flags & MXCASF_PERFORM_DNS_LOOKUP ) {
		perform_dns_lookup = TRUE;
	} else {
		perform_dns_lookup = FALSE;
	}

#if MXCA_CHANNEL_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', user '%s', host '%s'",
		fname, this->pv->PVname, this->username, this->hostname ));
#endif

	mxca_server = channel_pv->epicsServer;

	mx_record_list = mxca_server->mx_record_list;

	list_head = mx_get_record_list_head_struct( mx_record_list );

	if ( list_head->connection_acl == NULL ) {
		this->read_access_permitted = TRUE;
		this->write_access_permitted = TRUE;
	} else {
		/* See if we have already looked up this hostname. */

		mx_status = mx_hash_table_lookup_key( mxca_server->hash_table,
						this->hostname, &table_entry );

		if ( mx_status.code == MXE_SUCCESS ) {
			connection_allowed = *((mx_bool_type *) table_entry);
		} else
		if ( mx_status.code == MXE_NOT_FOUND ) {
			mx_status = mx_check_socket_connection_acl_permissions(
				mx_record_list, this->hostname,
				perform_dns_lookup, &connection_allowed );

			if ( mx_status.code != MXE_SUCCESS )
				return;

			key_value =
			    (mx_bool_type *) malloc( sizeof(mx_bool_type) );

			if ( key_value == NULL ) {
				(void) mx_error( MXE_OUT_OF_MEMORY, fname,
				"Ran out of memory trying to allocate "
				"an integer to store the connection status "
				"of '%s'.", this->hostname );

				return;
			}

			*key_value = connection_allowed;

#if MXCA_CHANNEL_DEBUG
			MX_DEBUG(-2,
			("%s: Inserting new hostname key '%s', value = %d",
			fname, this->hostname, (int) *key_value));
#endif
			mx_status = mx_hash_table_insert_key(
						mxca_server->hash_table,
						this->hostname, key_value );

			if ( mx_status.code != MXE_SUCCESS )
				return;
		} else {
			return;
		}

		if ( connection_allowed ) {
			this->read_access_permitted = TRUE;
			this->write_access_permitted = TRUE;
		} else {
			this->read_access_permitted = FALSE;
			this->write_access_permitted = FALSE;
		}
	}

	return;
}

mxcaChannel::~mxcaChannel( void )
{
#if MXCA_CHANNEL_DEBUG
	static const char fname[] = "mxcaChannel::~mxcaChannel()";

	MX_DEBUG(-2,("%s destructor: PV '%s', user '%s', host '%s'",
		fname, this->pv->PVname, this->username, this->hostname));
#endif

	return;
}

bool
mxcaChannel::readAccess( void ) const
{
	static const char fname[] = "mxcaChannel::readAccess()";

	mx_bool_type show_permission_tests = FALSE;

#if MXCA_CHANNEL_DEBUG
	show_permission_tests = TRUE;
#else
	unsigned long flags;

	flags = this->pv->epicsServer->mxca_server_flags;

	if ( flags & MXCASF_SHOW_PERMISSION_TESTS ) {
		show_permission_tests = TRUE;
	}
#endif

	if ( show_permission_tests ) {
		MX_DEBUG(-2,
		("%s: PV '%s', user '%s', host '%s', permitted = %d",
			fname, this->pv->PVname, this->username, this->hostname,
			this->read_access_permitted));
	}

	if ( this->read_access_permitted ) {
		return true;
	} else {
		return false;
	}
}

bool
mxcaChannel::writeAccess( void ) const
{
	static const char fname[] = "mxcaChannel::writeAccess()";

	mx_bool_type show_permission_tests = FALSE;

#if MXCA_CHANNEL_DEBUG
	show_permission_tests = TRUE;
#else
	unsigned long flags;

	flags = this->pv->epicsServer->mxca_server_flags;

	if ( flags & MXCASF_SHOW_PERMISSION_TESTS ) {
		show_permission_tests = TRUE;
	}
#endif

	if ( show_permission_tests ) {
		MX_DEBUG(-2,
		("%s: PV '%s', user '%s', host '%s', permitted = %d",
			fname, this->pv->PVname, this->username, this->hostname,
			this->write_access_permitted));
	}

	if ( this->write_access_permitted ) {
		return true;
	} else {
		return false;
	}
}

void
mxcaChannel::setOwner( const char * const user, const char * const host )
{
#if MXCA_CHANNEL_DEBUG
	static const char fname[] = "mxcaChannel::setOwner()";
#endif

	this->username = user;
	this->hostname = host;

#if MXCA_CHANNEL_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', user '%s', host '%s'",
		fname, this->pv->PVname, this->username, this->hostname));
#endif

	return;
}

