/*
 * Name:    mxcaMcaRecord.cpp
 *
 * Purpose: This file defines the mxcaMcaRecord class.
 *
 * Note: The MX record used by the mxcaPV:configure_mx_field_pv() method
 *       is actually the MX timer record for the object.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2012 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_MCA_DEBUG		TRUE

#include <iostream>
#include <fstream>
#include <cfloat>

#include "mxcaServer.h"
#include "mxcaMcaRecord.h"

extern "C" {
#include "mx_driver.h"
}

mxcaMcaRecord::mxcaMcaRecord( mxcaServer *epics_server,
				const char *epics_record_name ) :
					mxcaRecord( epics_server,
						epics_record_name,
						NULL )
{
#if MXCA_MCA_DEBUG
	static const char fname[] = "mxcaMcaRecord::mxcaMcaRecord()";
#endif

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for epics mca record '%s'",
		fname, epics_record_name));
#endif

	/* We start out with 0 MCA channels. */

	this->num_mca_channels = 0;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s complete for epics mca record '%s'",
		fname, epics_record_name));
#endif
}

/*----*/

mx_status_type
mxcaMcaRecord::add_mca( MX_RECORD *mx_mca_record )
{
#if MXCA_MCA_DEBUG
	static const char fname[] = "mxcaMcaRecord::add_mca()";
#endif
	mxcaMcaChannelStruct *mca_channel;
	int i;
	char epics_pv_name[200];

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s: MX mca '%s'", fname, mx_mca_record->name));
#endif
	i = this->num_mca_channels;

	if ( i >= MXCA_MCA_MAX_MCAS ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"Attempted to add too many MCA channels to "
		"EPICS MCA record '%s' at channel '%s'.  "
		"The maximum allowed is %d.",
			this->get_epics_record_name(),
			mx_mca_record->name,
			MXCA_MCA_MAX_MCAS );
	}

	mca_channel = &( this->mca_array[i] );

	this->num_mca_channels++;

	mca_channel->mx_record = mx_mca_record;
	mca_channel->channel_number = i;

	snprintf( epics_pv_name, sizeof(epics_pv_name), "mca%d.VAL", i );

	mca_channel->val_pv = new mxcaMcaChValPV( this, i, epics_pv_name );

	return MX_SUCCESSFUL_RESULT;
}

/*----*/

mxcaMcaChannelStruct *
mxcaMcaRecord::get_mca_channel( int channel_number )
{
	static const char fname[] = "mxcaMcaRecord::get_mca_channel()";

	mxcaMcaChannelStruct *mca_channel;

	if ( ( channel_number < 0 )
	  || ( channel_number >= this->num_mca_channels ) )
	{
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Requested MCA channel number %d for EPICS MCA '%s' "
		"is outside the allowed range of 0 to %d.",
			channel_number,
			this->get_epics_record_name(),
			this->num_mca_channels - 1 );

		return NULL;
	}

	mca_channel = &( this->mca_array[ channel_number ] );

	return mca_channel;
}

/*----*/

mx_status_type
mxcaMcaRecord::setup_global_pvs( int master_mca_channel_number )
{
	static const char fname[] = "mxcaMcaRecord::setup_global_pvs()";

	MX_RECORD *master_mx_mca_record;

	if ( this->num_mca_channels <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"No MX MCA channels were specified for EPICS MCA '%s'.",
			this->get_epics_record_name() );
	}

	/* The global PVs are attached to the specified master MCA channel. */

	if ( ( master_mca_channel_number < 0 )
	  || ( master_mca_channel_number >= this->num_mca_channels ) )
	{
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The requested master MX MCA channel number for EPICS MCA '%s' "
		"is outside the allowed range of 0-%d.",
			this->get_epics_record_name(),
			this->num_mca_channels - 1 );
	}

	master_mx_mca_record =
		this->mca_array[ master_mca_channel_number ].mx_record;

	this->mx_record = master_mx_mca_record;

	/* Create PVs that are for the detector system as a whole. */

	this->acquiring_pv = new mxcaMcaAcquiringPV( this );

	this->erase_all_pv = new mxcaMcaEraseAllPV( this );

	this->erase_start_pv = new mxcaMcaEraseStartPV( this );

	this->preset_live_pv = new mxcaMcaPresetLivePV( this );

	this->preset_real_pv = new mxcaMcaPresetRealPV( this );

	this->start_all_pv = new mxcaMcaStartAllPV( this );

	this->stop_all_pv = new mxcaMcaStopAllPV( this );

	return MX_SUCCESSFUL_RESULT;
}

/*-----------------------------------------------------------------------*/

mxcaMcaAcquiringPV::mxcaMcaAcquiringPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "Acquiring.VAL" )
{
#if MXCA_MCA_DEBUG
	static const char fname[] = "mxcaMcaAcquiringPV::mxcaMcaAcquiringPV()";
#endif

	mx_status_type mx_status;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	mx_status = mxcaPV::configure_mx_field_pv( "busy" );

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMcaEraseAllPV::mxcaMcaEraseAllPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "EraseAll.VAL" )
{
#if MXCA_MCA_DEBUG
	static const char fname[] =
			"mxcaMcaEraseStartPV::mxcaMcaEraseAllPV()";
#endif
	MX_RECORD *mx_timer_record;
	mx_status_type mx_status;

	mx_timer_record = this->mx_record;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s', MX timer '%s'",
		fname, this->PVname, mx_timer_record->name));
#endif

	mx_status = mxcaPV::configure_mx_field_pv( "clear" );

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMcaEraseStartPV::mxcaMcaEraseStartPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "EraseStart.VAL" )
{
#if MXCA_MCA_DEBUG
	static const char fname[] =
			"mxcaMcaEraseStartPV::mxcaMcaEraseStartPV()";
#endif
	MX_RECORD *mx_timer_record;
	mx_status_type mx_status;

	mx_timer_record = this->mx_record;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s', MX timer '%s'",
		fname, this->PVname, mx_timer_record->name));
#endif

	mx_status = mxcaPV::configure_mx_field_pv( "start" );

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMcaPresetLivePV::mxcaMcaPresetLivePV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "PresetLive.VAL" )
{
	static const char fname[] =
			"mxcaMcaPresetLivePV::mxcaMcaPresetLivePV()";

	mxcaMcaRecord *epics_mca_record;
	mxcaMcaChannelStruct *first_mca_channel;
	MX_RECORD *first_mx_mca_record;
	mx_status_type mx_status;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif
	epics_mca_record = (mxcaMcaRecord *) epics_record;

	/* Associate this PV with the first MX MCA record. */ 

	first_mca_channel = &( epics_mca_record->mca_array[0] );

	if ( first_mca_channel == (mxcaMcaChannelStruct *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaMcaChannelStruct pointer for the first MCA channel "
		"of EPICS record '%s' is NULL.",
			epics_record->get_epics_record_name() );

		return;
	}

	first_mx_mca_record = first_mca_channel->mx_record;

	if ( first_mx_mca_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for the first MCA channel "
		"of EPICS record '%s' is NULL.",
			epics_record->get_epics_record_name() );

		return;
	}

	this->mx_record = first_mx_mca_record;

	mx_status = mxcaPV::configure_mx_field_pv( "preset_live_time" );

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMcaPresetRealPV::mxcaMcaPresetRealPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "PresetReal.VAL" )
{
	static const char fname[] =
			"mxcaMcaPresetRealPV::mxcaMcaPresetRealPV()";

	mxcaMcaRecord *epics_mca_record;
	mxcaMcaChannelStruct *first_mca_channel;
	MX_RECORD *first_mx_mca_record;
	mx_status_type mx_status;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif
	epics_mca_record = (mxcaMcaRecord *) epics_record;

	/* Associate this PV with the first MX MCA record. */ 

	first_mca_channel = &( epics_mca_record->mca_array[0] );

	if ( first_mca_channel == (mxcaMcaChannelStruct *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaMcaChannelStruct pointer for the first MCA channel "
		"of EPICS record '%s' is NULL.",
			epics_record->get_epics_record_name() );

		return;
	}

	first_mx_mca_record = first_mca_channel->mx_record;

	if ( first_mx_mca_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for the first MCA channel "
		"of EPICS record '%s' is NULL.",
			epics_record->get_epics_record_name() );

		return;
	}

	this->mx_record = first_mx_mca_record;

	mx_status = mxcaPV::configure_mx_field_pv( "preset_real_time" );

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMcaStartAllPV::mxcaMcaStartAllPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "StartAll.VAL" )
{
#if MXCA_MCA_DEBUG
	static const char fname[] =
			"mxcaMcaStartAllPV::mxcaMcaStartAllPV()";
#endif
	MX_RECORD *mx_timer_record;
	mx_status_type mx_status;

	mx_timer_record = this->mx_record;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s', MX timer '%s'",
		fname, this->PVname, mx_timer_record->name));
#endif

	mx_status = mxcaPV::configure_mx_field_pv( "start" );

	return;
}

/*-----------------------------------------------------------------------*/

mxcaMcaStopAllPV::mxcaMcaStopAllPV( mxcaRecord *epics_record ) :
			mxcaPV( epics_record, "StopAll.VAL" )
{
#if MXCA_MCA_DEBUG
	static const char fname[] =
			"mxcaMcaStopAllPV::mxcaMcaStopAllPV()";
#endif
	MX_RECORD *mx_timer_record;
	mx_status_type mx_status;

	mx_timer_record = this->mx_record;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s', MX timer '%s'",
		fname, this->PVname, mx_timer_record->name));
#endif

	mx_status = mxcaPV::configure_mx_field_pv( "stop" );

	return;
}

/*=======================================================================*/

mxcaMcaChValPV::mxcaMcaChValPV( mxcaRecord *epics_record,
				int channel_number,
				const char *epics_pv_name ) :
					mxcaPV( epics_record, epics_pv_name )
{
#if MXCA_MCA_DEBUG
	static const char fname[] = "mxcaMcaChValPV::mxcaMcaChValPV()";
#endif
	mxcaMcaRecord *epics_mca_record;
	mxcaMcaChannelStruct *mca_channel;

	MX_RECORD *mx_mca_record;
	mx_status_type mx_status;

	epics_mca_record = (mxcaMcaRecord *) epics_record;

	mca_channel = epics_mca_record->get_mca_channel( channel_number );

	mx_mca_record = mca_channel->mx_record;

#if MXCA_MCA_DEBUG
	MX_DEBUG(-2,("%s: MCA channel %d for EPICS MCA '%s' is '%s', PV = '%s'",
		fname, channel_number,
		epics_mca_record->get_epics_record_name(),
		mx_mca_record->name, epics_pv_name ));
#endif
	this->mx_record = mx_mca_record;

	mx_status = mxcaPV::configure_mx_field_pv( "channel_array" );

	return;
}

caStatus
mxcaMcaChValPV::read( void )
{
	static const char fname[] = "mxcaMcaChValPV::read()";

	mxcaPV::read();

	MX_DEBUG(-2,("%s: mx_value = %p, epics_value = %p",
		fname, this->mx_value, this->epics_value));

	this->genCopy( this->mx_value );

	this->set_status_severity_and_timestamp( 0, 0 );

	this->value_changed_manual_override = TRUE;

	this->post_value_changed();

	return S_casApp_success;
}

