/*
 * Name:    mxcaServer.cpp
 *
 * Purpose: This file defines the mxcaServer class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2014 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_SERVER_DEBUG		FALSE

#define MXCA_SERVER_DEBUG_ATTACH	FALSE

#define MXCA_SERVER_DEBUG_EXIST_TEST	FALSE

#define MXCA_SERVER_DEBUG_FINDPV	FALSE

#define MXCA_SERVER_DEBUG_SCAN		FALSE

#include <iostream>
#include <fstream>

#include "mxcaServer.h"
#include "mxcaMotorRecord.h"

mxcaServer::mxcaServer ( const char *mx_database_filename,
			double scan_period,
			unsigned long mxca_server_flags,
			unsigned long mx_network_debug_flags,
			unsigned long epics_network_debug_flags ) : caServer()
{
	static const char fname[] = "mxcaServer::mxcaServer()";

	MX_LIST_HEAD *list_head;
	MX_PIPE *callback_pipe;
	MX_INTERVAL_TIMER *master_timer;
	MX_VIRTUAL_TIMER *callback_timer;
	mx_status_type mx_status;

#if MXCA_SERVER_DEBUG
	MX_DEBUG(-2,("%s invoked for database file '%s'",
		fname, mx_database_filename));
#endif

	this->mxca_server_flags = mxca_server_flags;

	/*** Create an empty MX database. ***/

	mx_status = mx_initialize_runtime();

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	mx_status = mx_initialize_drivers();

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	this->mx_record_list = mx_initialize_record_list();

	if ( this->mx_record_list == (MX_RECORD *) NULL ) {
		mx_status = mx_error( MXE_OUT_OF_MEMORY, fname,
		"Unable to create an MX record list." );
		
		exit(1);
	}

	/*
	 * Save a pointer to the mxcaServer object in the application_ptr
	 * element of the MX_LIST_HEAD structure for the list head record
	 * called "mx_database".
	 */

	list_head = mx_get_record_list_head_struct( this->mx_record_list );

	list_head->application_ptr = this;

	/*** Set the MX program name to 'mxca_server'. ***/

	strlcpy( list_head->program_name, "mxca_server",
				MXU_PROGRAM_NAME_LENGTH );

	/*** Associate the record list with an empty EPICS PV list. ***/

	mx_status = this->initializePVList();

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/*** Set the network debugging flags. ***/

	mx_multi_set_debug_flags( mx_record_list, mx_network_debug_flags );

	this->epics_network_debug_flags = epics_network_debug_flags;

	/*** Read in the database and initialize the hardware. ***/

	mx_status = mx_read_database_file( this->mx_record_list,
					mx_database_filename, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	mx_status = mx_finish_database_initialization( this->mx_record_list );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	mx_status = mx_initialize_hardware( this->mx_record_list, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/* Create a hash table to save DNS lookup information for
	 * the mxcaChannel class.
	 */

	mx_status = mx_hash_table_create( &(this->hash_table),
					MXU_HOSTNAME_LENGTH+1, 50, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/*** Turn on MX record processing. ***/

	mx_status = mx_initialize_database_processing( this->mx_record_list );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/*** Turn on alarm, value, and log events. ***/

	this->event_mask |= 
		( alarmEventMask() | valueEventMask() | logEventMask() );

	/*---------------------------------------------------------------*/

	/*** Setup the MX callback pipe ***/

	mx_status = mx_setup_callback_pipe( this->mx_record_list,
						&callback_pipe );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/*** Setup the MX master timer with a 100 millisecond period. ***/

	double master_timer_period = 0.1;

	mx_status = mx_virtual_timer_create_master( &master_timer,
							master_timer_period );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	list_head->master_timer = master_timer;

	/* The callback timer will need a message object to write to the
	 * callback pipe telling the main thread that it is time to scan
	 * the EPICS database.  We must malloc() that structure here,
	 * since it is unsafe to do it in the context of the callback
	 * function
	 */

	this->poll_callback_message =
		(MX_CALLBACK_MESSAGE *) malloc( sizeof(MX_CALLBACK_MESSAGE) );

	if ( this->poll_callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		(void) mx_error( MXE_OUT_OF_MEMORY, fname,
		"The attempt to allocate memory for an MX_CALLBACK_MESSAGE "
		"structure failed." );

		exit(1);
	}

	this->poll_callback_message->callback_type = MXCBT_POLL;
	this->poll_callback_message->list_head = list_head;

	/* Setup the callback virtual timer that periodically writes a
	 * callback message through the MX callback pipe which will ask
	 * the main thread to scan the database.
	 *
	 * The timer is intentionally a one-shot timer, so the function
	 * mxca_server_scan_database() is expected to restart the timer
	 * at the beginning of that function.  This ensures that there
	 * is, at most, one poll callback message in the callback pipe
	 * at any given time and prevents the callback pipe from filling
	 * up with a multitude of poll callbacks if the callback function
	 * is slow.
	 */

	mx_status = mx_virtual_timer_create( &callback_timer,
				(MX_INTERVAL_TIMER *) list_head->master_timer,
				MXIT_ONE_SHOT_TIMER,
				mx_request_value_changed_poll,
				this->poll_callback_message );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	list_head->callback_timer = callback_timer;

	this->scan_period = scan_period;

	/* Create list_head->server_callback_handle_table by invoking
	 * mx_initialize_callback_support().
	 */

	mx_status = mx_initialize_callback_support( this->mx_record_list );

	if ( mx_status.code != MXE_SUCCESS )
		exit(1);

	/*---------------------------------------------------------------*/

#if MXCA_SERVER_DEBUG
	MX_DEBUG(-2,("%s complete for database file '%s'",
		fname, mx_database_filename));
#endif
}

pvExistReturn
mxcaServer::pvExistTest ( const casCtx &ctx, const char *pPVName )
{
#if MXCA_SERVER_DEBUG_EXIST_TEST
	static const char fname[] = "mxcaServer::pvExistTest()";
#endif

	mxcaPV *pv;

	pv = this->findPV( pPVName );

	if ( pv == NULL ) {

#if MXCA_SERVER_DEBUG_EXIST_TEST
		MX_DEBUG(-2,("%s: returning pverDoesNotExistHere for '%s'",
			fname, pPVName));
#endif

		return pvExistReturn( pverDoesNotExistHere );
	} else {

#if MXCA_SERVER_DEBUG_EXIST_TEST
		MX_DEBUG(-2,("%s: returning pverExistsHere for '%s'",
			fname, pPVName));
#endif

		return pvExistReturn( pverExistsHere );
	}
}

pvAttachReturn
mxcaServer::pvAttach ( const casCtx &ctx, const char *pvname )
{
#if MXCA_SERVER_DEBUG_ATTACH
	static const char fname[] = "mxcaServer::pvAttach()";
#endif

	mxcaPV *pv;

	pv = this->findPV( pvname );

	if ( pv == NULL ) {

#if MXCA_SERVER_DEBUG_ATTACH
		MX_DEBUG(-2,("%s: returning S_casApp_pvNotFound for '%s'",
			fname, pvname));
#endif

		return pvAttachReturn( S_casApp_pvNotFound );
	} else {

#if MXCA_SERVER_DEBUG_ATTACH
		MX_DEBUG(-2,("%s: returning '%s' for '%s'",
			fname, pv->PVname, pvname));
#endif

		return pvAttachReturn( *pv );
	}
}

/*=================== Private member functions ====================*/

mxcaPV *
mxcaServer::findPV ( const char *requested_pvname )
{
#if MXCA_SERVER_DEBUG_FINDPV
	static const char fname[] = "mxcaServer::findPV()";
#endif

	mxcaPV *this_pv, *list_head_pv;
	const char *this_pvname;
	char val_pvname[512];
	const char *pvname_ptr;

	if ( strchr( requested_pvname, '.' ) != NULL ) {
		pvname_ptr = requested_pvname;

#if MXCA_SERVER_DEBUG_FINDPV
		MX_DEBUG(-2,("%s invoked for PV name '%s'",
			fname, pvname_ptr ));
#endif
	} else {
		snprintf( val_pvname, sizeof(val_pvname),
			"%s.VAL", requested_pvname );

		pvname_ptr = val_pvname;

#if MXCA_SERVER_DEBUG_FINDPV
		MX_DEBUG(-2,("%s invoked for PV name '%s' converted to '%s'",
			fname, requested_pvname, pvname_ptr ));
#endif
	}

	list_head_pv = this->PVList;

	if ( list_head_pv == NULL ) {
		return NULL;
	}

	this_pv = list_head_pv->nextPV;

	while ( this_pv != list_head_pv ) {

		this_pvname = this_pv->PVname;

		if ( strcmp( pvname_ptr, this_pvname ) == 0 ) {

#if MXCA_SERVER_DEBUG_FINDPV
			MX_DEBUG(-2,
			("%s: match found for '%s', this_pv = %p",
				fname, pvname_ptr, this_pv));
#endif

			break;
		}

		this_pv = this_pv->nextPV;
	}

	if ( this_pv == list_head_pv ) {
		this_pv = NULL;
	}

#if MXCA_SERVER_DEBUG_FINDPV
	MX_DEBUG(-2,("%s: returning this_pv = %p", fname, this_pv));
#endif

	return this_pv;
}

mx_status_type
mxcaServer::initializePVList ( void )
{
#if MXCA_SERVER_DEBUG
	static const char fname[] = "mxcaServer::initializePVList()";
#endif

	MX_RECORD *mx_record, *mx_record_list;

#if MXCA_SERVER_DEBUG
	MX_DEBUG(-2,("%s invoked", fname ));
#endif

	/* Create an mxcaRecord object corresponding to the MX
	 * list head record.
	 */

	mx_record_list = this->mx_record_list;

	mx_record = mx_get_record( mx_record_list, "mx_database" );

	this->epicsRecordList = new mxcaRecord( this, "", mx_record );

	/* Create an empty PV corresponding to the list head. */

	this->PVList = new mxcaPV( this, "@@@.@@@" );

	/* The PV list is a circular linked list. */

	this->PVList->previousPV = this->PVList;
	this->PVList->nextPV = this->PVList;

	return MX_SUCCESSFUL_RESULT;
}

void
mxcaServer::start( void )
{
#if MXCA_SERVER_DEBUG
	static const char fname[] = "mxcaServer::start()";

	MX_DEBUG(-2,("%s invoked, scan_period = %g",
		fname, this->scan_period ));
#endif
	MX_LIST_HEAD *list_head;

	list_head = mx_get_record_list_head_struct( this->mx_record_list );

	(void) mx_virtual_timer_start(
	    (MX_VIRTUAL_TIMER *) list_head->callback_timer, this->scan_period );
}

void
mxcaServer::scan( MX_VIRTUAL_TIMER *mx_callback_timer )
{
	static const char fname[] = "mxcaServer::scan()";

	mxcaPV *PVList, *pv;
	mx_status_type mx_status;

#if MXCA_SERVER_DEBUG_SCAN
	MX_DEBUG(-2,("****** %s invoked ******", fname));
#endif
	PVList = this->PVList;

	if ( PVList == (mxcaPV *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The PVList pointer for the MX Channel Access server is NULL."
		"  This will prevent the server from scanning PVs." );

		return;
	}

	pv = PVList->nextPV;

	if ( pv == (mxcaPV *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The first PV on the PVList is NULL.  "
		"This will prevent the server from scanning PVs." );

		return;
	}

	/* Restart the callback virtual timer.  This is done here to make sure
	 * that the timer is restarted, since a failure in pv->event_ready()
	 * might otherwise prevent the timer from being restarted.
	 */

	mx_status = mx_virtual_timer_restart( mx_callback_timer );

	/* Scan through the PVs. */

	while ( pv != PVList ) {

		if ( pv->value_is_monitored ) {
#if MXCA_SERVER_DEBUG_SCAN
			MX_DEBUG(-2,("%s: PV '%s' value_is_monitored = %d",
				fname, pv->getName(),
				(int) pv->value_is_monitored ));
#endif
			pv->check_for_event();
		}

		if ( pv->nextPV == (mxcaPV *) NULL ) {
			(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"pv->nextPV for PV '%s' is NULL.  "
			"The current PV scan will stop at this point.",
				pv->PVname );

			break;		/* Exit the while() loop. */
		}

		pv = pv->nextPV;
	}

#if MXCA_SERVER_DEBUG_SCAN
	MX_DEBUG(-2,("****** %s complete ******", fname));
#endif

	return;
}

/*-------------------------------------------------------------------------*/

